-- Copyright 2011 Stephan Hennig
--
--  This program is free software: you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation, either version 3 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program.  If not, see <http://www.gnu.org/licenses/>.


--- Helper module for checking the Google Books Corpus 2009-07-15.
--
-- @author Stephan Hennig
-- @copyright 2012 Stephan Hennig
-- @release version 0.2
--
-- @trick Prevent LuaDoc from looking past here for module description.
--[[ Trick LuaDoc into entering 'module' mode without using that command.
module(...)
--]]


-- Import external modules.
local handler_prototype = require('modules.GoogleBooks-2009-07-15.handler_prototype')
local unicode = require('unicode')


-- Module table.
local M = {}


-- Short-cuts.
local Umatch = unicode.utf8.match


--- Load totalcounts file.
-- Load data from totalcounts file and store it for later usage.  Do
-- some validity checks on totalcounts file during loading.
--
-- @param path  Path to corpus files.
-- @param tag  String determining a particular corpus.
-- @return Table with data read from totalcounts file.
local function __load_totalcounts_file(path, tag)
   io.stderr:write('opening totalcounts file\n')
   local fname = path .. 'googlebooks-' .. tag .. '-totalcounts-20090715.txt'
   local f = assert(io.open(fname , 'r'))
   -- Skip first line with table head.
   f:lines()()
   local total_counts = {}
   local minyear = math.huge
   local maxyear = -math.huge
   -- Read data from totalcounts file.
   for line in f:lines() do
      local year, match_ct, page_ct, volume_ct = string.match(line, '^(%d+)\t(%d+)\t(%d+)\t(%d+)$')
      year = tonumber(year)
      match_ct = tonumber(match_ct)
      page_ct = tonumber(page_ct)
      volume_ct = tonumber(volume_ct)
      -- Check for valid fields.
      if not (year > 1500 and year < 2100 and
              match_ct >= page_ct and page_ct >= volume_ct and volume_ct >= 1) then
         io.stderr:write('Inconsistent data in totalcounts file: ', line, '\n')
      end
      -- Check if year appears at most once in totalcounts file.
      if total_counts[year] then
         io.stderr:write('Non-unique year in totalcounts file: ', line, '\n')
      end
      -- Store data.
      total_counts[year] = {
         match_ct = match_ct,
         page_ct = page_ct,
         volume_ct = volume_ct,
      }
      -- Calculate minimum/maximum years.
      if year < minyear then minyear = year end
      if year > maxyear then maxyear = year end
   end
   f:close()
   io.stderr:write('year range: (', minyear, ', ', maxyear, ')\n')
   return total_counts
end


--- Parse a line in the corpus.
-- Decomposes a line in the Google Books 2009-07-15 corpus into the
-- provided fields.
--
-- @param line  A text line from the corpus.
-- @return ngram  An ngram.
-- @return year  The year.
-- @return match_ct  Number of matches.
-- @return page_ct  Number of pages.
-- @return volume_ct  Number of books.
local function __parse_line(line)
   -- Catch entries.
   local ngram, year, match_ct, page_ct, volume_ct = Umatch(line, '^([^\t]+)\t(%d+)\t(%d+)\t(%d+)\t(%d+)$')
   -- Convert to number.
   year = tonumber(year)
   match_ct = tonumber(match_ct)
   page_ct = tonumber(page_ct)
   volume_ct = tonumber(volume_ct)
   return ngram, year, match_ct, page_ct, volume_ct
end


--- Check a corpus file.
--
-- @param path  Path to corpus files.
-- @param pat  A pattern string determining file(s) to open.
-- @param tag  String determining a particular corpus.
-- @param ngrams  Table of unique ngrams found so far.
-- @param annual  Table of cummulated annual matches.
-- @return Minimum match, page, volume frequencies found in the file.
-- Frequencies do not necessarily correspond to the same ngram.
local function __check_file(path, pat, tag, ngrams, annual)
   local lastngram = nil
   local match_cum = 40
   local page_cum = 40
   local volume_cum = 40
   local match_min = math.huge
   local page_min = math.huge
   local volume_min = math.huge
   local years
   local fname = path .. 'googlebooks-' .. tag .. '-1gram-20090715-' .. pat .. '.csv.zip'
   io.stderr:write('opening ', fname, '\n')
   local f = assert(io.popen('unzip -p ' .. fname))
   for line in f:lines() do
      local ngram, year, match_ct, page_ct, volume_ct = __parse_line(line)
      -- Check for valid ngram.
      if ngram == nil or ngram == '' then
         io.stderr:write('invalid ngram: ', line, '\n')
      end
      -- Check for valid counts.
      if not (match_ct >= page_ct and page_ct >= volume_ct and volume_ct >= 1) then
         io.stderr:write('invalid count: ', line, '\n')
      end
      -- New ngram?
      if ngram ~= lastngram then
         -- Check for valid volume count of last ngram.
         if volume_cum < 40 then
            io.stderr:write('invalid ngram count: ngram: ', lastngram, '  volumes: ', volume_cum, '\n')
         end
         -- New minimum frequency?
         if match_cum < match_min then match_min = match_cum end
         if page_cum < page_min then page_min = page_cum end
         if volume_cum < volume_min then volume_min = volume_cum end
         -- Check for unique ngram sequence.
         if ngrams[ngram] then
            io.stderr:write('no consecutive ngram sequence: ', line, '\n')
         end
         -- Remember current ngram.
         lastngram = ngram
         -- Mark ngram as seen.
         ngrams[ngram] = true
         -- Reset ngram counts.
         match_cum = 0
         page_cum = 0
         volume_cum = 0
         -- Provide new table of matches per year.
         years = {}
      end
      -- Unique ngram/year combination?
      if years[year] then
         io.stderr:write('no unique ngram/year combination: ', line, '\n')
      end
      -- Mark ngram/year combination as seen.
      years[year] = true
      -- Cummulate annual matches.
      annual[year] = (annual[year] or 0) + match_ct
      -- Cummulate ngram counts.
      match_cum = match_cum + match_ct
      page_cum = page_cum + page_ct
      volume_cum = volume_cum + volume_ct
   end
   f:close()
   io.stderr:write('min ', match_min, ',', page_min, ',', volume_min, '\n')
   if volume_min < 40 then
      io.stderr:write('invalid corpus file: minimum volume count < 40!\n')
   end
   return match_min, page_min, volume_min
end


--- Check cummulated annual counts in corpus against data from
--- totalcounts file.
--
-- @param annual  Table of cummulated annual matches.
-- @param annual  Table of annual tables of cummulated counts.
local function __check_counts(total_counts, annual)
   -- Check for all indices (years) from table total_counts are present
   -- in table annual.
   for year, _ in pairs(total_counts) do
      if not annual[year] then
         io.stderr:write('Missing year in corpus: ', year, '\n')
      end
   end
   -- Check for all indices (years) from table annual are present in
   -- table total_counts.
   for year, _ in pairs(annual) do
      if not total_counts[year] then
         io.stderr:write('Missing year in totalcounts file: ', year, '\n')
      end
   end
   -- Check for valid counts in corpus.
   for year, _ in pairs(annual) do
      local tc = total_counts[year]
      if annual[year] > total_counts[year].match_ct then
         io.stderr:write('annual count mismatch: year ', year, '\n')
         io.stderr:write('totalcounts:   ', tc.match_ct, ' ', tc.page_ct, ' ', tc.volume_ct, '\n')
         io.stderr:write('annual counts: ', an.match_ct, ' ', an.page_ct, ' ', an.volume_ct, '\n')
      end
   end
end


--- Check validity of the corpus.
--
-- @param path  Path to corpus files.
-- @param tag  String determining a particular corpus.
-- @return Number of unique ngrams in corpus.
local function check(path, tag)
   local ngrams = {}
   local annual = {}
   local match_min = math.huge
   local page_min = math.huge
   local volume_min = math.huge
   local handler = handler_prototype:new()

   -- Iterate over all files in the corpus.
   for pat in handler:corpus_patterns() do
      local m,p,v = __check_file(path, pat, tag, ngrams, annual)
      if m < match_min then match_min = m end
      if p < page_min then page_min = p end
      if v < volume_min then volume_min = v end
   end
   io.stderr:write('corpus minima matches,pages,volumes = ', match_min, ',', page_min, ',', volume_min, '\n')

   local total_counts = __load_totalcounts_file(path, tag)
   __check_counts(total_counts, annual)

   -- Count number of unique ngrams.
   local c = 0
   for _, __ in pairs(ngrams) do
      c = c + 1
   end
   io.stderr:write(c, ' unique ngrams found in corpus\n')
end
M.check = check


-- Return module table.
return M
