-- Copyright 2013 Stephan Hennig
--
--  This program is free software: you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation, either version 3 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program.  If not, see <http://www.gnu.org/licenses/>.


--- Helper module for checking the Google Books Corpus 2012-07-01.
--
-- @author Stephan Hennig
-- @copyright 2012 Stephan Hennig
-- @release version 0.2
--
-- @trick Prevent LuaDoc from looking past here for module description.
--[[ Trick LuaDoc into entering 'module' mode without using that command.
module(...)
--]]


-- Import external modules.
local handler_prototype = require('modules.GoogleBooks-2012-07-01.handler_prototype')
local unicode = require('unicode')


-- Module table.
local M = {}


-- Short-cuts.
local Umatch = unicode.utf8.match


--- Load totalcounts file.
-- Load data from totalcounts file and store it for later usage.  Do
-- some validity checks on totalcounts file during loading.
--
-- @param path  Path to corpus files.
-- @param tag  String determining a particular corpus.
-- @return Table with data read from totalcounts file.
local function __load_totalcounts_file(path, tag)
   io.stderr:write('opening totalcounts file\n')
   local fname = path .. 'googlebooks-' .. tag .. '-totalcounts-20120701.txt'
   local f = assert(io.open(fname , 'r'))
   local total_counts_file = f:read('*a')
   f:close()
   local total_counts = {}
   local minyear = math.huge
   local maxyear = -math.huge
   -- Read data from totalcounts file.
   for line in string.gmatch(total_counts_file, '(%d+)\t(%d+)\t(%d+)\t(%d+)') do
      local year, match_ct, page_ct, volume_ct = string.match(line, '^(%d+)\t(%d+)\t(%d+)\t(%d+)$')
      year = tonumber(year)
      match_ct = tonumber(match_ct)
      page_ct = tonumber(page_ct)
      volume_ct = tonumber(volume_ct)
      -- Check for valid fields.
      if year <= 1500 or year >= 2100 or
         match_ct <= 0 or page_ct <= 0 or volume_ct <= 0 or
         match_ct < page_ct or page_ct < volume_ct then
         io.stderr:write('Inconsistent data in totalcounts file: ', line, '\n')
      end
      -- Check if year appears at most once in totalcounts file.
      if total_counts[year] then
         io.stderr:write('Non-unique year in totalcounts file: ', line, '\n')
      end
      -- Store data.
      total_counts[year] = {
         match_ct = match_ct,
         page_ct = page_ct,
         volume_ct = volume_ct,
      }
      -- Calculate minimum/maximum years.
      if year < minyear then minyear = year end
      if year > maxyear then maxyear = year end
   end
   io.stderr:write('year range: (', minyear, ', ', maxyear, ')\n')
   return total_counts
end


--- Table that maps valid POS tags to `true`.
-- This table is used to verify validity of tags found in ngrams.
--
-- @class table
-- @name __is_POS_tag
local __is_POS_tag = {

   ['ADP'] = true,
   ['ADJ'] = true,
   ['VERB'] = true,
   ['NOUN'] = true,
   ['NUM'] = true,
   ['PRON'] = true,
   ['ADV'] = true,
   ['CONJ'] = true,
   ['DET'] = true,
   ['PRT'] = true,
   ['X'] = true,

}


--- Extract POS data from ngram.
-- Strings with POS tags are of the form <code>ngram_tag</code>.  POS
-- tags are optional.  This function extracts POS tags if they are
-- present.  Validity of found POS tags is checked via table
-- `__is_pos_tag`.
--
-- @param ngram  An ngram, optionally annotated with POS data.
-- @return Sanitized ngram.
-- @return Extracted pos tag or `nil`.
local function __extract_pos(ngram)
   -- Try to separate ngram and tag.
   local cngram, ctag = Umatch(ngram, '^(.+)_([A-Z]+)$')
   -- Valid tag found?
   if ctag and __is_POS_tag[ctag] then return cngram, ctag end
   -- Ngram doesn't contain POS data.
   return ngram, nil
end


--- Parse a line in the corpus.
-- Decomposes a line in the Google Books 2012-07-01 corpus into the
-- provided fields.
--
-- @param line  A text line from the corpus.
-- @return ngram  An ngram.
-- @return year  The year.
-- @return match_ct  Number of matches.
-- @return volume_ct  Number of books.
local function __parse_line(line)
   -- Catch entries.
   local ngram, year, match_ct, volume_ct = Umatch(line, '^([^\t]+)\t(%d+)\t(%d+)\t(%d+)$')
   -- Convert to number.
   year = tonumber(year)
   match_ct = tonumber(match_ct)
   volume_ct = tonumber(volume_ct)
   return ngram, year, match_ct, volume_ct
end


--- Check a corpus file.
--
-- @param path  Path to corpus files.
-- @param pat  A pattern string determining file(s) to open.
-- @param tag  String determining a particular corpus.
-- @param ngrams  Table of unique ngrams found so far.
-- @param annual  Table of cummulated annual matches.
-- @return Minimum match, volume frequencies found in the file.
-- Frequencies do not necessarily correspond to the same ngram.
local function __check_file(path, pat, tag, ngrams, annual)
   local lastngram = nil
   local match_cum = 40
   local volume_cum = 40
   local match_min = math.huge
   local volume_min = math.huge
   local years
   local first_char = {}
   local first_alpha = {}
   local first_digit = {}
   local first_alnum = {}
   local fname = path .. 'googlebooks-' .. tag .. '-1gram-20120701-' .. pat .. '.gz'
   io.stderr:write('opening ', fname, '\n')
   local f = assert(io.popen('gunzip -c ' .. fname))
   for line in f:lines() do
      local ngram, year, match_ct, volume_ct = __parse_line(line)
      local plain_ngram, tag = __extract_pos(ngram)
      -- Check for valid ngram.
      if plain_ngram == nil or plain_ngram == '' then
         io.stderr:write('invalid ngram: ', line, '\n')
      end
      -- Check for valid counts.
      if not (match_ct >= volume_ct and volume_ct >= 1) then
         io.stderr:write('invalid count: ', line, '\n')
      end
      -- New ngram?
      if ngram ~= lastngram then
         -- Check for valid volume count of last ngram.
         if volume_cum < 40 then
            io.stderr:write('invalid ngram count: ngram: ', lastngram, '  volumes: ', volume_cum, '\n')
         end
         -- New minimum frequency?
         if match_cum < match_min then match_min = match_cum end
         if volume_cum < volume_min then volume_min = volume_cum end
         -- Check for unique ngram sequence.
         if ngrams[ngram] then
            io.stderr:write('no consecutive ngram sequence: ', line, '\n')
         end
         -- Store first character of ngram.
         first_char[Umatch(ngram, '^(.)')] = true
         -- Store first letter character of plain ngram.
         local ch = Umatch(plain_ngram, '(%a)')
         if ch then first_alpha[ch] = true end
         -- Store first digit character of plain ngram.
         local ch = Umatch(plain_ngram, '(%d)')
         if ch then first_digit[ch] = true end
         -- Store first alpha-numeric character of plain ngram.
         local ch = Umatch(plain_ngram, '(%w)')
         if not ch then
            io.stderr:write('no alpha-numeric character in ngram: ', line, '\n')
         else
            first_alnum[ch] = true
         end
         -- Remember current ngram.
         lastngram = ngram
         -- Mark ngram as seen.
         ngrams[ngram] = true
         -- Reset ngram counts.
         match_cum = 0
         volume_cum = 0
         -- Provide new table of matches per year.
         years = {}
      end
      -- Unique ngram/year combination?
      if years[year] then
         io.stderr:write('no unique ngram/year combination: ', line, '\n')
      end
      -- Mark ngram/year combination as seen.
      years[year] = true
      -- Cummulate annual matches.
      annual[year] = (annual[year] or 0) + match_ct
      -- Cummulate ngram counts.
      match_cum = match_cum + match_ct
      volume_cum = volume_cum + volume_ct
   end
   f:close()
   -- Write first characters to file.
   local f = assert(io.open(path .. 'first_char.' .. pat, 'w'))
   for ch,_ in pairs(first_char) do
      f:write(ch, '\n')
   end
   f:close()
   -- Write first letter characters to file.
   local f = assert(io.open(path .. 'first_alpha.' .. pat, 'w'))
   for ch,_ in pairs(first_alpha) do
      f:write(ch, '\n')
   end
   f:close()
   -- Write first digit characters to file.
   local f = assert(io.open(path .. 'first_digit.' .. pat, 'w'))
   for ch,_ in pairs(first_digit) do
      f:write(ch, '\n')
   end
   f:close()
   -- Write first alpha-numeric characters to file.
   local f = assert(io.open(path .. 'first_alnum.' .. pat, 'w'))
   for ch,_ in pairs(first_alnum) do
      f:write(ch, '\n')
   end
   f:close()
   io.stderr:write('min ', match_min, ',', volume_min, '\n')
   if volume_min < 40 then
      io.stderr:write('invalid corpus file: minimum volume count < 40!\n')
   end
   return match_min, volume_min
end


--- Check cummulated annual counts in corpus against data from
--- totalcounts file.
--
-- @param annual  Table of cummulated annual matches.
-- @param annual  Table of annual tables of cummulated counts.
local function __check_counts(total_counts, annual)
   -- Check that all indices (years) from table total_counts are present
   -- in table annual.
   for year, _ in pairs(total_counts) do
      if not annual[year] then
         io.stderr:write('Missing year in corpus: ', year, '\n')
      end
   end
   -- Check that all indices (years) from table annual are present in
   -- table total_counts.
   for year, _ in pairs(annual) do
      if not total_counts[year] then
         io.stderr:write('Missing year in totalcounts file: ', year, '\n')
      end
   end
   -- Check for valid counts in corpus.
   for year, _ in pairs(annual) do
      local tc = total_counts[year]
      if annual[year] > total_counts[year].match_ct then
         io.stderr:write('annual count mismatch: year ', year, '\n')
         io.stderr:write('totalcounts:   ', tc.match_ct, ' ', tc.page_ct, ' ', tc.volume_ct, '\n')
         io.stderr:write('annual counts: ', an.match_ct, ' ', an.volume_ct, '\n')
      end
   end
end


--- Check validity of the corpus.
--
-- @param path  Path to corpus files.
-- @param tag  String determining a particular corpus.
-- @return Number of unique ngrams in corpus.
local function check(path, tag)
   local ngrams = {}
   local annual = {}
   local match_min = math.huge
   local volume_min = math.huge
   local handler = handler_prototype:new()

   -- Iterate over all files in the corpus.
   for pat in handler:corpus_patterns() do
      local m,v = __check_file(path, pat, tag, ngrams, annual)
      if m < match_min then match_min = m end
      if v < volume_min then volume_min = v end
   end
   io.stderr:write('corpus minima matches,volumes = ', match_min, ',', volume_min, '\n')

   local total_counts = __load_totalcounts_file(path, tag)
   __check_counts(total_counts, annual)

   -- Count number of unique ngrams.
   local c = 0
   for _, __ in pairs(ngrams) do
      c = c + 1
   end
   io.stderr:write(c, ' unique ngrams found in corpus\n')
end
M.check = check


-- Return module table.
return M
