-- Copyright 2013 Stephan Hennig
--
--  This program is free software: you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation, either version 3 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program.  If not, see <http://www.gnu.org/licenses/>.


--- Helper module for handling the Google Books Corpus 2012-07-01.
--
-- @author Stephan Hennig
-- @copyright 2012 Stephan Hennig
-- @release version 0.2
--
-- @trick Prevent LuaDoc from looking past here for module description.
--[[ Trick LuaDoc into entering 'module' mode without using that command.
module(...)
--]]


-- Import external modules.
local handler_base_prototype = require('modules.handler_base_prototype')
local unicode = require('unicode')


-- Prototype class table, derived from basic prototype class.
local M = handler_base_prototype:new()


-- Short-cuts.
local Tinsert = table.insert

local Ugmatch = unicode.utf8.gmatch
local Umatch = unicode.utf8.match


-- String determining the particular Google Books corpus to read
-- (contains language and ngram number).
local tag = nil
M.tag = tag


-- Table containing patterns covering all files in the corpus.
local patterns = {}
-- '0' to '9'
for i=48,57 do
   Tinsert(patterns, string.char(i))
end
-- 'a' to 'z'
for i=97,122 do
   Tinsert(patterns, string.char(i))
end
Tinsert(patterns, 'other')
Tinsert(patterns, 'punctuation')
M.patterns = patterns


--- Open a file from the corpus.
--
-- @param self  Callee reference.
-- @param pat  A pattern string determining file(s) to open.
-- @return File handle.
local function open_file(self, pat)
   local path = self.path
   if not (type(path) == 'string') then
      error('Invalid file path: ' .. tostring(path))
   end
   if not string.find(path, '[\\/]$', -1) then
      path = path .. '/'
   end
   local tag = self.tag
   if not (type(tag) == 'string') then
      error('Invalid corpus tag: ' .. tostring(tag))
   end
   local fname = path .. 'googlebooks-' .. tag .. '-1gram-20120701-' .. pat .. '.gz'
   io.stderr:write('opening ', fname, '\n')
   local f = assert(io.popen('gunzip -c ' .. fname))
   return f
end
M.open_file = open_file


--- Generator factory: get all records in corpus.
--
-- @param self  Callee reference.
-- @return Record generator.
local function corpus_records(self)
   -- Create corpus chunk generator.
   local next_chunk = self:corpus_chunks()
   -- Corpus chunk.
   local chunk
   -- Create empty record match generator.
   local next_record = function () end
   -- Get filter function.
   local is_valid = self.record_filter
   if not (is_valid == nil or type(is_valid) == 'function') then
      error('Invalid record filter.  Must be nil or a function, found: ' .. tostring(is_valid))
   end
   return function ()
      local ngram, year, match_ct, volume_ct
      -- Filter loop.
      repeat
         -- Get next record.
         ngram, year, match_ct, volume_ct = next_record()
         -- End of chunk?
         if not ngram then
            chunk = nil
            next_record = nil
            -- Time for a garbage collection.
            chunk = next_chunk()
            -- End of corpus?
            if not chunk then return end
            -- Create new record match generator.
            next_record = Ugmatch(chunk, '(.-)\t(.-)\t(.-)\t(.-)\n')
            ngram, year, match_ct, volume_ct = next_record()
         end
         -- Convert to number.
         year = tonumber(year)
         match_ct = tonumber(match_ct)
         volume_ct = tonumber(volume_ct)
      until (is_valid == nil) or is_valid(ngram, year, match_ct, volume_ct)
      return ngram, year, match_ct, volume_ct
   end
end
M.corpus_records = corpus_records


--- Table that maps valid POS tags to `true`.
-- This table is used to verify validity of tags found in ngrams.
--
-- @class table
-- @name __is_POS_tag
local __is_POS_tag = {

   ['ADP'] = true,
   ['ADJ'] = true,
   ['VERB'] = true,
   ['NOUN'] = true,
   ['NUM'] = true,
   ['PRON'] = true,
   ['ADV'] = true,
   ['CONJ'] = true,
   ['DET'] = true,
   ['PRT'] = true,
   ['X'] = true,

}


--- Extract POS data from ngram.
-- Strings with POS tags are of the form <code>ngram_tag</code>.  POS
-- tags are optional.  This function extracts POS tags if they are
-- present.  Validity of found POS tags is checked via table
-- `__is_pos_tag`.
--
-- @param ngram  An ngram, optionally annotated with POS data.
-- @return Sanitized ngram.
-- @return Extracted pos tag or `nil`.
local function __extract_pos(ngram)
   -- Try to separate ngram and tag.
   local cngram, ctag = Umatch(ngram, '^(.+)_([A-Z]+)$')
   -- Valid tag found?
   if ctag and __is_POS_tag[ctag] then return cngram, ctag end
   -- Ngram doesn't contain POS data.
   return ngram, nil
end


--- Generator factory: get all ngrams in corpus.
--
-- @param self  Callee reference.
-- @return Ngram generator.
local function corpus_ngrams(self)
   -- Create corpus record generator.
   local next_record = self:corpus_records()
   local ngram, year, match_ct, volume_ct = next_record()
   -- Get filter function.
   local is_valid = self.ngram_filter
   if not (is_valid == nil or type(is_valid) == 'function') then
      error('Invalid ngram filter.  Must be nil or a function, found: ' .. tostring(is_valid))
   end
   return function ()
      local curr_ngram, curr_match_ct, curr_volume_ct
      -- Filter loop.
      repeat
         -- Check if there's still an ngram on the shelf.
         if not ngram then return nil end
         -- Initialize current ngram.
         curr_ngram = ngram
         curr_match_ct, curr_volume_ct = 0, 0
         -- Cummulate counts until we find a different ngram.
         repeat
            curr_match_ct = curr_match_ct + match_ct
            curr_volume_ct = curr_volume_ct + volume_ct
            ngram, year, match_ct, volume_ct = next_record()
         until ngram ~= curr_ngram
         local curr_tag
         curr_ngram, curr_tag = __extract_pos(curr_ngram)
      until (is_valid == nil) or is_valid(curr_ngram, curr_match_ct, curr_volume_ct, curr_tag)
      return curr_ngram, curr_match_ct, curr_volume_ct, curr_tag
   end
end
M.corpus_ngrams = corpus_ngrams


-- Return prototype class.
return M
