-- Copyright 2011 Stephan Hennig
--
--  This program is free software: you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation, either version 3 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program.  If not, see <http://www.gnu.org/licenses/>.


--- Helper module for frequency class related calculations.
--
-- @author Stephan Hennig
-- @copyright 2013 Stephan Hennig
-- @release version 0.2
--
-- @trick Prevent LuaDoc from looking past here for module description.
--[[ Trick LuaDoc into entering 'module' mode without using that command.
module(...)
--]]


-- Module table.
local M = {}


-- Short-cuts.
local log = math.log
local exp = math.exp
local floor = math.floor
local log2 = math.log(2)


-- Parameter for generalization of frequency class calculation.
local offset = 0.5-- see http://de.wikipedia.org/wiki/H%C3%A4ufigkeitsklasse


--- Calculate frequency class of a number.
--
-- @param ref  Reference frequency.
-- @param frequency  Frequency of which to calculate frequency class.
-- @return Frequency class relative to reference frequency.
local function frequency_class(ref, frequency)
  return floor(log(ref / frequency) / log2 + offset)
end
M.frequency_class = frequency_class


--- Print a range of frequencies and corresponding classes.
--
-- @param ref  Reference frequency.
-- @param frequency  Base frequency.
-- @param eps  Epsilon that controls the range of frequencies to show.
local function print_frequency_range(ref, frequency, eps)
   for i = eps, -eps, -1 do
      io.stderr:write(string.format('%2d  %15s  %3d\n', i, tostring(frequency+i), frequency_class(ref, frequency+i)))
   end
end
M.print_frequency_range = print_frequency_range


local minfreq_class_mismatch = 0
local minfreq_not_a_class_boundary = 0


--- Print number of errors encountered during minimum frequency
--- calculations.
--
local function print_minfreq_errors()
   io.stderr:write('class mismatch errors: ', error_class_mismatch, '\n')
   io.stderr:write('not a class boundary errors: ', error_not_a_class_boundary, '\n')
end
M.print_minfreq_errors = print_minfreq_errors


--- Calculate lower limit of a certain frequency class.
--
-- @param ref  Reference frequency.
-- @param class  Frequency class.
-- @return Minimum frequency of the given frequency class.
local function minimum_frequency(ref, class)
  -- This calculation can occasionally be off by 1 for negative
  -- frequency classes, i.e., large frequencies.  Therefore, we do some
  -- consistency checks and apply simple default corrections in case.
  local minfreq = floor(ref / exp((class + 1 - offset) * log2)) + 1
  -- Check frequency class of calculated minimum frequency.
  local check_class = frequency_class(ref, minfreq)
  if check_class ~= class then
     minfreq_class_mismatch = minfreq_class_mismatch + 1
     -- Recalculate, applying default correction.
     check_class = frequency_class(ref, minfreq + 1)
     -- Still off?
     if check_class ~= class then
        io.stderr:write('Error calculating minimum frequency: Wrong class, frequency off by more than 1. ')
        io.stderr:write(string.format('ref=%d, class=%d, minfreq=%d\n', ref, class, minfreq))
        print_frequency_range(ref, minfreq, 3)
     end
     minfreq = minfreq + 1
  end
  -- Check if we're really at a frequency class boundary.
  local check_class_next = frequency_class(ref, minfreq - 1)
  if check_class_next ~= check_class + 1 then
     minfreq_not_a_class_boundary = minfreq_not_a_class_boundary + 1
     -- Recalculate, applying default correction.
     check_class_next = frequency_class(ref, minfreq - 2)
     --Still off?
     if check_class_next ~= check_class + 1 then
        io.stderr:write('Error calculating minimum frequency: Not at class boundary, off by more than 1. ')
        io.stderr:write(string.format('ref=%d, class=%d, minfreq=%d\n', ref, class, minfreq))
        print_frequency_range(ref, minfreq, 3)
     end
     minfreq = minfreq - 1
  end
  return minfreq
end
M.minimum_frequency = minimum_frequency


--- Print a table of frequency classes and their minimum frequencies.
--
-- @param ref  Reference frequency.
-- @param from  Lowest frequency class to print.
local function print_class_table(ref, from)
   -- Calculate maximum frequency class to print.
   local to = frequency_class(ref, 1) - 1
   print(string.format('frequency class   minimum frequency'))
   for class = from,to do
      local minfreq = minimum_frequency(ref, class)
      print(string.format('%15d   %10d', class, minfreq))
   end
end
M.print_class_table = print_class_table


-- Return module table.
return M
