-- Copyright 2011 Stephan Hennig
--
--  This program is free software: you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation, either version 3 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program.  If not, see <http://www.gnu.org/licenses/>.


--- Base prototype class for handling corpora.
--
-- @author Stephan Hennig
-- @copyright 2012 Stephan Hennig
-- @release version 0.2
--
-- @trick Prevent LuaDoc from looking past here for module description.
--[[ Trick LuaDoc into entering 'module' mode without using that command.
module(...)
--]]


-- Import external modules.
local frequencyclass = require('modules.frequencyclass')
local unicode = require('unicode')


-- Prototype class table.
local M = {}


-- Short-cuts.
local Tinsert = table.insert

local Ugmatch = unicode.utf8.gmatch


--- Handle access to elements not implemented in prototype class.
-- This triggers an error.
local function __not_implemented()
   error('Invalid access to an element not implemented in prototype class!')
end


-- Meta table making all access to the original table invalid.
local __meta_not_implemented = {
   __index = __not_implemented,
   __newindex = __not_implemented,
}


-- Number determining buffer size for (piped) file reading.
local buffersize = 4096
M.buffersize = buffersize


-- A string determining the output format.
-- Default output format is a plain list of ngrams.
local output_format = 'n'
M.output_format = output_format


-- String determining path to the corupus file(s) to read.
local path = nil
M.path = path


-- Table containing patterns covering all files in the corpus.
local patterns = {}
setmetatable(patterns, __meta_not_implemented)
M.patterns = patterns


-- A number representing the reference frequency for frequency
-- calculations in case the reference ngram cannot be found in the
-- corpus.
local reference_frequency = nil
M.reference_frequency = reference_frequency


-- A string representing the reference ngram for frequency calculations.
-- In case the reference ngram can be found in the corpus, this variable
-- dominates reference frequency.
local reference_ngram = nil
M.reference_ngram = reference_ngram


--- Constructor.
--
-- @param self  Callee reference.
-- @param o  Object to be used for initialisation (optional).
-- @return A new base corpus object.
local function new(self, o)
   o = o or {}
   --- Initialize corpus table.
   --
   -- @class table
   -- @name corpus
   --
   -- @field ngram  Table, mapping indices to ngrams.
   -- @field frequency  Table, mapping ngrams to frequencies.
   -- @field max_frequency  Number, representing maximum frequency of an
   -- ngram in the corpus.
   o.corpus = {
      ngram = {},
      frequency = {},
      max_frequency = 0,
   }
   setmetatable(o, self)
   self.__index = self
   return o
end
M.new = new


--- Open a file from the corpus.
--
-- @param self  Callee reference.
-- @param pat  A pattern string determining file(s) to open.
-- @return File handle.
local function open_file(self, pat)
   __not_implemented()
end
M.open_file = open_file


--- Generator factory: get all file patterns in corpus.
--
-- @param self  Callee reference.
-- @return File pattern generator.
local function corpus_patterns(self)
   -- Table of file patterns.
   local t = self.patterns
   if type(t) == 'string' then
      t = { t }
   end
   if not (type(t) == 'table') then
      error('Invalid file patterns: ' .. tostring(t))
   end
   local i = 0
   return function ()
      i = i + 1
      return t[i]
   end
end
M.corpus_patterns = corpus_patterns


--- Generator factory: get all files in corpus.
--
-- @param self  Callee reference.
-- @return File handle generator.
local function corpus_files(self)
   -- Create corpus pattern generator.
   local next_pattern = self:corpus_patterns()
   return function ()
      -- In every file iteration, do a garbage collection cycle.
      do
         local c1 = collectgarbage('count')
         collectgarbage('collect')
         local c2 = collectgarbage('count')
         io.stderr:write('collect garbage: ', c1, ' > ', c2, '\n')
      end
      local pat = next_pattern()
      if pat then
         return self:open_file(pat)
      end
   end
end
M.corpus_files = corpus_files


--- Generator factory: get all chunks in corpus.
-- Chunked reading can reduce corpus processing time significantly (25
-- to 30 percent).
--
-- @param self  Callee reference.
-- @return Chunk generator.
local function corpus_chunks(self)
   -- Create corpus file handle generator.
   local next_file = self:corpus_files()
   -- File handle.
   local f = next_file()
   -- Get local reference to buffer size.
   local BUFFERSIZE = self.buffersize
   return function ()
      -- Read next chunk.
      local chunk, rest = f:read(BUFFERSIZE, '*l')
      -- EOF?
      if not chunk then
         f:close()
         -- Open next corpus file.
         f = next_file()
         if not f then return end
         chunk, rest = f:read(BUFFERSIZE, '*l')
      end
      if rest then chunk = chunk .. rest .. '\n' end
      return chunk
   end
end
M.corpus_chunks = corpus_chunks


--- Generator factory: get all lines in corpus.
--
-- @param self  Callee reference.
-- @return Line generator.
local function corpus_lines(self)
   -- Create corpus chunk generator.
   local next_chunk = self:corpus_chunks()
   -- Corpus chunk.
   local chunk
   -- Create empty line match generator.
   local next_line = function () end
   return function ()
      -- Read next line.
      local line = next_line()
      -- End of chunk?
      if not line then
         chunk = nil
         next_line = nil
         -- Time for a garbage collection.
         chunk = next_chunk()
         -- End of corpus?
         if not chunk then return end
         -- Create new line match generator.
         next_line = Ugmatch(chunk, '(.-)\n')
         line = next_line()
      end
      return line
   end
end
M.corpus_lines = corpus_lines


--- Generator factory: get all records in corpus.
--
-- @param self  Callee reference.
-- @return Record generator.
local function corpus_records(self)
   __not_implemented()
end
M.corpus_records = corpus_records


--- Generator factory: get all ngrams in corpus.
--
-- @param self  Callee reference.
-- @return Ngram generator.
local function corpus_ngrams(self)
   __not_implemented()
end
M.corpus_ngrams = corpus_ngrams


--- Read a corpus.
-- Frequencies of identical ngrams are cummulated.
--
-- @param self  Callee reference.
-- @return A table with fields `ngram`, `frequency`, `max_frequency`.
-- Field `ngram` is a table, which maps indices to ngrams.  Field
-- `frequency` is a table, which maps ngrams to cummulated frequencies.
-- Field `max_frequency` is a number, containing maximum frequency of an
-- ngram in the corpus.
local function read_corpus(self)
   -- Map indices to ngrams.  Via this table, ngrams can be proccessed
   -- in read order.
   local ngram = self.corpus.ngram
   -- Map ngrams to (cummulated) frequencies.
   local frequency = self.corpus.frequency
   -- Maximum frequency.
   local max_frequency = 0
   -- Create ngram generator.
   local next_ngram = self:corpus_ngrams()
   for ng, f in next_ngram do
      local f_known = frequency[ng]
      -- New ngram?
      if not f_known then
         Tinsert(ngram, ng)
         f_known = 0
      end
      -- Cummulate frequency.
      f_known = f_known + f
      frequency[ng] = f_known
      -- New frequency maximum?
      if f_known > max_frequency then max_frequency = f_known end
   end
   self.corpus.max_frequency = max_frequency
   return self.corpus
end
M.read_corpus = read_corpus


--- Output corpus as a list of ngrams.
-- The list is output to stdout.
--
-- @param self  Callee reference.
local function output_list_ngram(self)
   for _,n in ipairs(self.corpus.ngram) do
      io.stdout:write(n, '\n')
   end
end
M.output_list_ngram = output_list_ngram


--- Output corpus as a list 'ngram frequency'.
-- The list is output to stdout.
--
-- @param self  Callee reference.
local function output_list_ngram_frequency(self)
   local frequency = self.corpus.frequency
   for _,n in ipairs(self.corpus.ngram) do
      io.stdout:write(n, ' ', frequency[n], '\n')
   end
end
M.output_list_ngram_frequency = output_list_ngram_frequency


--- Output corpus as a list 'ngram frequencyclass'.
-- The list is output to stdout.
--
-- @param self  Callee reference.
local function output_list_ngram_frequencyclass(self)
   local frequency = self.corpus.frequency
   local ref_n = self.reference_ngram
   if not type(ref_n) == 'string' then
      error('Invalid reference ngram: ' .. tostring(ref_n))
   end
   local ref_f = frequency[ref_n]
   if not ref_f then
      io.stderr:write('reference ngram not found in corpus, trying reference frequency\n')
      ref_f = self.reference_frequency
   end
   if not type(ref_f) == 'number' then
      error('Invalid reference frequency: ' .. tostring(ref_f))
   end
   for _,n in ipairs(self.corpus.ngram) do
      io.stdout:write(n, ' ', frequencyclass.frequency_class(ref_f, frequency[n]), '\n')
   end
end
M.output_list_ngram_frequencyclass = output_list_ngram_frequencyclass


--- Output corpus in a particular list format.
-- The list is output to stdout.
--
-- @param self  Callee reference.
local function output_corpus(self)
   local output_format = self.output_format
   if output_format == nil then
      -- Do nothing.
   elseif output_format == 'n' then
      self:output_list_ngram()
   elseif output_format == 'nf' then
      self:output_list_ngram_frequency()
   elseif output_format == 'nc' then
      self:output_list_ngram_frequencyclass()
   else
      error('Invalid output format: ' .. tostring(output_format))
   end
end
M.output_corpus = output_corpus


-- Return prototype class.
return M
